#!/perl

package Scraps;
use Scraps::Errors;
use Scraps::Util;
use DBI;
use Data::Dumper;
use XML::ValidWriter;
use XML::Doctype;
require XML::Simple;

use constant TRUE => 1;
use constant FALSE => 0;

=pod 

=head1 NAME 

Scraps - Server-side implementation of the Scraps API

=head1 DESCRIPTION

This is an implementation of the Scraps API for information exchange and
retrieval.

This backend provides commands that have the following signature:
	foo(username,database_context,[ function args ])
 where
  username => a validated username.
  database_context => a database context for whatever databasing system
 					    is being used.  In this case, DBI, but this calling
						signature should be preservable with other databases
  [ function args ] => the other arguments passed to the XML-RPC call.

=head1 PUBLIC INTERFACE

=item B<newScrap($username,$db,$data)>

This function takes the supplied scrap data (in $data) and inserts
it into the database as a new scrap, with a newly generated ID.

It returns the full data of the newly created scrap if successful.  
Otherwise, the appropriate error condition is returned.

=cut 

sub newScrap 
{
	my $username = shift;
	my $db = shift;
	my $data = shift;

	## Validate data
	my ($rc,$message) = validate_new_scrap($data);
	print STDERR "Recieved $rc, $message from validate_new_scrap()\n";
	if (SUCCESS != $rc) {
		print STDERR "Invalid data!\n";
		return(INVALID_DATA,"TODO: Put Real Error Here");
	}

	##Generate scrap ID.	
	my $scrap_id = uuidgen();

	## Insert scrap into scraps table
	$sth = $db->prepare(qq{
		INSERT INTO scrap (id,title,date_created) VALUES (?,?,NOW())});
	## TODO: Error check this!
	$sth->execute($scrap_id, $data->{title});
	$sth->finish;

	## TODO: Error check this!
	store_scrap_data(
		$db, 
		$scrap_id,
		{
			data => $data->{data}->{data},
			datatype => $data->{data}->{type},
			description => $data->{description},
			creator_name => $data->{creator}->{name},
			creator_email => $data->{creator}->{email},
		}
	);

	## Store keywords
	## TODO: Error check this!
	store_scrap_keywords(
		$db,
		$scrap_id,
		$data->{keywords});

	## Note that we have no contributor info for a new scrap. 
	return fetchScrap($username,$db,$scrap_id);
}

=pod 

=item B<deleteScrap($username,$db,$scrap_id)>

Deletes the scrap with the given $scrap_id from the database.
Returns ID_NOT_EXIST if the id does not exist on the server.

=cut

sub deleteScrap 
{
	my $username = shift;
	my $db = shift;
	my $scrap_id = shift;


	my $sth = $db->prepare("SELECT id from scrap where id = ?");
	$sth->execute($scrap_id);
	my $ary = $sth->fetchrow_arrayref();
	if(!defined($ary)) {
		return(ID_NOT_EXIST,"That ID does not exist on the server");
	}
	$sth->finish();

	## Delete from the scraps table
	$sth = $db->prepare(qq{DELETE FROM scrap where id = ?});
	$sth->execute($scrap_id);
	$sth->finish;

	## Delete from the contributor table
	$sth = $db->prepare(qq{DELETE FROM contributor where id = ?});
	$sth->execute($scrap_id);
	$sth->finish;

	return(SUCCESS,TRUE);
}


=pod 

=item B<fetchScrap($username,$db,$scrap_id)>

Fetches the scrap with id $scrap_id.  Returns all scrap data on success.
Otherwise, returns the appropriate error condition.

=cut

sub fetchScrap 
{
	my $username = shift;
	my $db = shift;
	my $scrap_id = shift;

	print STDERR "Received scrap_id [$scrap_id]\n";
	my $sth = $db->prepare(
	qq{
		SELECT	s.title,
				s.creator_name,
				s.creator_email,
				s.date_created,
				s.date_modified,
				s.date_accessed,
				s.date_imported,
				s.description,
				s.datatype,
				s.data
		FROM scrap s
		WHERE 
			s.id = ? 
	});
	$sth->execute($scrap_id);

	my $ary = $sth->fetchrow_arrayref();
	if(!defined($ary)) {
		return(ID_NOT_EXIST,"That ID does not exist on the server");
	}
	$sth->finish();
	
	my $rv = {
		id		=> $scrap_id,
		title	=> $ary->[0],
		creator => {
			name	=> $ary->[1],
			email	=> $ary->[2]
		},
		date => {
			created  => $ary->[3],
			modified => $ary->[4],
			accessed => $ary->[5],
			imported => $ary->[6],
		},
		description => $ary->[7],
		data => {
			type => $ary->[8],
			data => $ary->[9]
		},
		contributor => []
	};	

	## Fetch keyword data
	$sth = $db->prepare(
		qq{
			SELECT
				keywords
			FROM scrap
			WHERE
				id = ?
		}
	);
	## TODO: Check return val for errors!
	$sth->execute($scrap_id);
	$row = $sth->fetchrow_arrayref();
	$sth->finish;
	
	## Remove leading/trailing ;;'s, as split() will cause
	## empty keywords to be generated if these are present.
	$row->[0] =~ s/^;;(.+);;$/$1/;
	foreach my $k (split(/;;/,$row->[0])) {	
		push @{$rv->{keywords}}, $k;
	}

	## Fetch contributor data
	$sth = $db->prepare(
		qq{
			SELECT	
				name,
				email,
				contrib_date,
				data
			FROM contributor
			WHERE
				scrap_id = ?
		}
	);
	## TODO: error check the database call!
	$sth->execute($scrap_id);

	$h = $sth->fetchall_arrayref({});
	## Add each contributor to the db.
	foreach my $c (@$h) {	
		push @{$rv->{contributor}}, {
			name => $c->{name},
			email => $c->{email},
			date => $c->{contrib_date},
			note => $c->{data}
		};
	}

	## Now, set date_accessed to now, and return $rv.  We want $rv->date->accessed
	## to have the previous access time, not this one.
	## TODO: error check the database call!
	$db->do(qq{UPDATE scrap SET date_accessed = NOW() WHERE id = '$scrap_id'});
	return(SUCCESS,$rv);
}

=pod 

=item B<exportScrap($username,$db,$scrap_id)>

Fetches the scrap with id $scrap_id and returns the scrap as exported XML
conforming to the Scraps DTD.
Otherwise, returns the appropriate error condition.

=cut

sub exportScrap 
{
	my $username = shift;
	my $db = shift;
	my $scrap_id = shift;
	my $no_emit_decl = shift || 0;

	my $scrap_xml;

	my ($rc,$scrap_data) = fetchScrap($username,$db,$scrap_id);
	if(SUCCESS != $rc) {
		## $scrap contains error message in this case.
		return($rc,$scrap_data);
	}

	print STDERR Dumper $scrap_data;

	## TODO: make scrapbook.dtd a config variable!
	my $dtd =  XML::Doctype->new('scrap',SYSTEM_ID => 'scrapbook.dtd');
	my $writer = XML::ValidWriter->new(
		DOCTYPE => $dtd,
		OUTPUT => \$scrap_xml
	);
	$writer->setDataMode(1);
	
	## This is here so that we don't emit the declaration when beign called by
	## exportSearch.
	$writer->xmlDecl unless $no_emit_decl;
	$writer->startTag('scrap', id => $scrap_data->{id});
	$writer->dataElement('title', ($scrap_data->{title} || ''));
	$writer->startTag('creator');
	  $writer->dataElement('name', ($scrap_data->{creator}->{name} || ''));
	  $writer->dataElement('email', $scrap_data->{creator}->{email} || '');
	$writer->endTag('creator');
	foreach my $cont (@{$scrap_data->{contributor}}) {	
		$writer->startTag('contributor');
		$writer->dataElement('name',$cont->{name} || '');
		$writer->dataElement('email',$cont->{email} || '');
		$writer->dataElement('date',$cont->{date} || '');
		if(defined($cont->{note})) {
			$writer->dataElement('note',$cont->{note});
		}
		$writer->endTag('contributor');
	}
	$writer->dataElement('description', $scrap_data->{description});
	foreach my $kw (@{$scrap_data->{keywords}}) {	
		$writer->dataElement('keyword',$kw);
	}
	foreach my $key (keys %{$scrap_data->{date}}) {
		next if (!defined($scrap_data->{date}->{$key}));
		$writer->dataElement('date', $scrap_data->{date}->{$key} || '', type => $key);
	}
	$writer->dataElement('data', $scrap_data->{data}->{data} || '', type => $scrap_data->{data}->{type});
	$writer->endTag('scrap');

	return(SUCCESS,$scrap_xml);
}

=pod 

=item B<saveScrap($username,$db,$scrap_id,$data)>

Updates scrap identified by $scrap_id with the information in $data.
if $scrap_id is an empty string, a new scrap is created from the given
data.  

Returns the saved scrap on success, or the appropriate error condition on 
failure.

=cut

sub saveScrap 
{
	my $username = shift;
	my $db = shift;
	my $scrap_id = shift;
	my $scrap = shift;

	my $add_data;

	if($scrap_id ne '') {
		## FIXME: Right now, we don't use this groups stuff
		## We can only mess with this scrap if:
		##	a) the scrap is not associated with an access group, or
		##	b) the scrap is associated with an access group, and the current user
		##	   is part of that group.
		## Note that this query is not optimized at all.  Ugh.
		my $sth = $db->prepare(
		qq{
			SELECT	s.title
			FROM scrap s, scrap_group sg, group_assoc g 
			WHERE 
				s.id = ? 
				AND (
					NOT(s.id = sg.scrap_id) 
					OR (s.id = sg.scrap_id AND sg.groupname = g.groupname AND g.username = ?)
				)
		});
		## TODO: error check the database call!
		$sth->execute($scrap_id,$username);

		if(!defined($sth->fetchrow_arrayref())) {
			$sth->finish();
			return(ID_NOT_EXIST,"That ID does not exist!");
		}
		$sth->finish();

		$add_data = {
			data => $scrap->{data}->{data},
			datatype => $scrap->{data}->{type},
			description => $scrap->{description},
			title => $scrap->{title},
		};

	} else {
		## Here, we were given a blank scrap ID.  So, we need to check that:
		## a) there is a scrap ID attached to the scrap in $scrap
		## b) that scrap ID doesn't exist on the server.
		## if both these conditions are met, we insert this scrap ID and then
		## proceed as an update.
		if(!exists($scrap->{'id'}) || length($scrap->{'id'}) != 32) {
			print STDERR "Scrap ID is $scrap->{id} \n";
			print STDERR "Scrap ID length is " .  length($scrap->{'id'}) . "\n";
			return(INVALID_DATA,"Need to give a scrap ID");
		}

		my $sth = $db->prepare(
		qq{SELECT title
			FROM scrap
			WHERE id = ?});
		## TODO: error check the database call!
		$sth->execute($scrap->{id});

		if(defined($sth->fetchrow_arrayref())) {
			$sth->finish();
			print STDERR "Scrap ID already exists!\n";
			return(ID_EXISTS,"That ID already exists!");
		}
		$sth->finish();

		## Otherwise, add it, and proceed as normal.
		## Insert scrap into scraps table
		$sth = $db->prepare(qq{
			INSERT INTO scrap (id,title,date_imported) VALUES (?,?,NOW())});
		## TODO: Error check this!
		$sth->execute($scrap->{id}, $scrap->{title});
		$sth->finish;
		
		$scrap_id = $scrap->{id};
		$add_data = {
			data => $scrap->{data}->{data},
			datatype => $scrap->{data}->{type},
			description => $scrap->{description},
			title => $scrap->{title},
			creator_name => $scrap->{creator}->{name},
			creator_email => $scrap->{creator}->{email},
		};
	}

	## TODO Validate data
#	my ($rc,$message) = validate_saved_scrap($data);
#	print STDERR "Recieved $rc, $message from validate_new_scrap()\n";
#	if ( $rc ) {
#		print STDERR "Invalid data!\n";
#		return(INVALID_DATA,"TODO: Put Real Error Here");
#	}

	## TODO: Check for errors.
	store_scrap_data(
		$db, 
		$scrap_id,
		$add_data
	);

	## Store keywords
	## TODO: check return for error.
	store_scrap_keywords(
		$db,
		$scrap_id,
		$scrap->{keywords});

	## Now update the contributor table.
	## We will just loop through all the contributors given us, and 
	## set the values.	
	## TODO: How do we do this easily?  We want to set it if not exists, 
	## and replace it if it does.
	if(defined($scrap->{contributor})) {
		print Dumper $scrap->{contributor};
		## FIXME: Check for error.
		store_scrap_contributors( $db, $scrap_id, $scrap->{contributor});
	}

	return fetchScrap($username,$db,$scrap_id);
}

=pod 

=item B<search($username,$db,$search)>

Searches the database for scraps matching the given search criteria 
(in $search).  

Returns an array containing anonymous hashes of:
		Scrap ID
		Title
		Description
		Last modified data
A valid but unsuccessful search returns an empty array.  If an error
is encountered, the appropriate error condition is returned.

=cut

sub search 
{
	my $username = shift;
	my $db = shift;
	my $s = shift;
	my ($rc, $clause, $message);

	($rc, $message) = validate_search($s);
	if(SUCCESS != $rc) {
		return($rc,$message);
	}

	return(INVALID_SEARCH,"Invalid search") unless ($s->{query});
	my $type = (keys %{$s->{query}})[0];

	($rc,$clause) = build_search_clause($s->{query});
	if(SUCCESS != $rc) {
		return(INVALID_SEARCH,$clause);
	}

	my $sql = q{
		SELECT
			id,
			title,
			IF(LENGTH(description) > 40, CONCAT(LEFT(description,40),' ...'),description) AS description,
			date_modified
		FROM scrap
		WHERE };
	$sql .= $clause;

	my $sth = $db->prepare($sql);
	$sth->execute();
	my $res = $sth->fetchall_arrayref({});

	if(!defined($res)) {
		return(SUCCESS,[]);
	} else {
		return(SUCCESS,$res);
	}
	
}

##
##  Internal function used to generate a search clause for the
##  sql database.
##
sub build_search_clause 
{
	my $query = shift;
	## Remove the <server> section if present... we're ignoring for now
	delete($query->{server}) if exists($query->{server});

	## See if we've got too many top-level operators.
	my ($count,$op);	
	foreach my $ant (qw(and or not)) {
		if(exists($query->{$ant})) {
			++$count;
			$op = $ant;
		}
	}
	if($count != 1) {
		return(INVALID_SEARCH,"Have $count boolean operators in query, expected 1.");
	}

	## Otherwise, we recursively build a query for $op.
	$qstring = build_query($op,$query->{$op});
	print STDERR "Built [$qstring] from $op\n";
	return (SUCCESS,$qstring);
}


##
##  Builds the query string appropriately.
##
sub build_query 
{
	my $op = shift;
	my $query = shift;

	my @args;
	foreach my $obj (@{$query}) {
		if (exists($obj->{keyword})) {
			push @args, "(keywords LIKE '%;;$obj->{keyword};;%')";
		}
		elsif (exists($obj->{'or'})) {
		}
		elsif (exists($obj->{'and'})) {
		}
		elsif (exists($obj->{'not'})) {
		}
		elsif (exists($obj->{'created'})) {
		}
		elsif (exists($obj->{'modified'})) {
		}
		elsif (exists($obj->{'accessed'})) {
		}
		elsif (exists($obj->{'imported'})) {
		}
	}
	return join(" $op ",@args);
}

=pod

=item B<exportSearch($username,$db,$search)>

Searches the database for scraps matching the given search criteria 
(in $search).  Returns the full data for these scraps, encoded as XML 
conforming to the Scraps DTD.

If an error is encountered, the appropriate error condition is returned.

=cut

sub exportSearch
{
	my $username = shift;
	my $db = shift;
	my $s = shift;

	my($rc,$res) = search($username,$db,$s);

	if(SUCCESS != $rc) {
		return($rc,$res);
	}

	my $export;
	foreach my $line (@$res) {
		my ($rc,$scrap_ex) = exportScrap($username,$db,$line->{id},1);
		if(SUCCESS != $rc) {
			return(INTERNAL_ERROR,"Error fetching data from database");
		}
		$export .= $scrap_ex;
	}

	return(SUCCESS,qq{<?xml version="1.0"?>\n<scrapbook>\n$export</scrapbook>\n});
}

=pod

=item B<importScrapbook($username,$db,$xml_text)>

Imports XML-formatted scraps into the database.

=cut

sub importScrapbook
{
	my $username = shift;
	my $db = shift;
	my $xml = shift || return(INVALID_DATA,"No XML data given!");	

	print STDERR Dumper $xml;

	## Parse the XML and create in-memory scraps.
	my $xs = new XML::Simple();
	my $ref = $xs->XMLin($xml, keyattr => [ 'type' ], forcearray => ['scrap']);
	my @status;

	print STDERR Dumper $ref;

	foreach my $scrap (@{$ref->{scrap}}) {
		## Munge name of data element into place, and fix spacing issues.
		$scrap->{data}->{data} = $scrap->{data}->{content};
		delete($scrap->{data}->{content});

		$scrap->{data}->{data} =~ s/^\s*(.+)\s*$/$1/;

		## Fix 'keyword' -> 'keywords'	
		$scrap->{keywords} = $scrap->{keyword};
		delete($scrap->{keyword});

		## Fix 'date' subelements.
		foreach my $date (keys %{$scrap->{date}}) {
			my $foo = $scrap->{date}->{$date}->{content};
			delete($scrap->{date}->{$date}->{content});
			$scrap->{date}->{$date} = $foo;
		}

		my ($rc,$new_scrap) = saveScrap($username,$db,'',$scrap);
		if(SUCCESS == $rc) {
			push (@status, { title => $scrap->{title}, id => $scrap->{id}, status => 'added' } );
		} 
		elsif(ID_EXISTS == $rc) {
			push (@status, { title => $scrap->{title}, id => $scrap->{id}, status => 'exists' } );
			print STDERR "Error:  ID_EXISTS:  $new_scrap\n";
		}
		else {
			push (@status, { title => $scrap->{title}, id => $scrap->{id}, status => 'invalid' } );
			print STDERR "Error:  $rc:  $new_scrap\n";
		}
	}

	print STDERR Dumper \@status;

	return(SUCCESS,\@status);
}

=pod

=back

=head1 INTERNAL FUNCTIONS

There are several internal functions used to support these operations.
They are documented here as not being documented.  =)  Do not use
them unless you know what you  are doing

=cut

##
##
##
sub validate_search
{
	my $search = shift;
	## TODO: write validation code.
	return SUCCESS;
}

##
## Validates data given for a new scrap.
##
sub validate_new_scrap 
{
	my $data = shift;
	## 1) ensure that all required fields are present
	## 2) ensure that all required and optional fields contain correct
	##    data (for some value of 'correct')
	## TODO: better data validation.  

	if(!defined($data->{data}) || !$data->{data}->{data}) {
		return(INVALID_DATA,"No data present");
	}
	if ( (!defined($data->{data}) || !$data->{data}->{type}) 
		 || $data->{data}->{type} !~ /(text|url|query)/  ) {
		return(INVALID_DATA,"invalid data type");
	}
	if (!defined($data->{creator}) || !$data->{creator}->{email}) {
		return(INVALID_DATA,"Invalid creator email");
	}
	if (!defined($data->{creator}) || !$data->{creator}->{name}) {
		return(INVALID_DATA,"Invalid creator name");
	}
	if (!$data->{title}) {
		return(INVALID_DATA,"Invalid title");
	}
	if (!$data->{description}) {
		return(INVALID_DATA,"Invalid description");
	}
	if (!defined($data->{keywords}) || !$data->{keywords}->[0]) {
		return(INVALID_DATA,"Invalid keywords");
	}
	
	return(SUCCESS,TRUE);
}

##
## store given scrap data.
##
sub store_scrap_data 
{
	my $db = shift;
	my $id = shift;
	my $parms = shift;
	my $sql = "UPDATE scrap SET ";
	my @setvars;
	## Add other fields as necessary.
	foreach my $i (qw(title creator_name creator_email description datatype data)) {
		if(exists($parms->{$i})) {
			print STDERR "store_scrap_data() got value [$parms->{$i}] for $i\n";
			$sql .= "$i = ?, ";
			push @setvars, $parms->{$i};
		}
	}
	$sql .= "date_modified = NOW() WHERE id = ?";
	push @setvars, $id;
	my $sth = $db->prepare($sql);
	if(!defined($sth)) {
		print STDERR "Didn't create statement handle!\n";
	}

	## TODO: errorcheck.
	my $rv = $sth->execute(@setvars);
	if($rv == 0) {
		print STDERR "Error executing SQL: ", $db->errstr;
	}
	$sth->finish;
}

## Store the keywords for the scrap.  
## TODO: Return a success/fail value!
sub store_scrap_keywords
{
	my $db = shift;
	my $scrap_id = shift;
	my $keywords = shift; ## Array ref

	## Join with special chars, and ensure each word is bounded on
	## both sides by the ;; characters.
	my $words = join(';;',@$keywords);
	$words = ";;$words;;";	

	my $sth = $db->prepare(
		qq{ UPDATE scrap SET keywords = ? WHERE id = ? }
	);
	$sth->execute($words, $scrap_id);
	$sth->finish;
}

##
##  Loop through all contributors, and:
##	a) if they exist, replace with new info
##  b) if they don't exist, add them.
##  Again, because of laziness, we will delete the old, then add the new
##  See above for rationale/explanation.
##

=pod

=item B<store_scrap_contributors($db,$contributors)>

=cut

sub store_scrap_contributors
{
	my $db = shift;
	my $scrap_id = shift;
	my $contributors = shift; ## Hashref for contributors.

#	## Possibilities;
#	## 1) adding new, in which case, no contributors.  Ignore.
#	## 2) editing, in which case we should only get the info that's 
#	##    changed?  Or do we get it all?
#
#	my $sth = $db->prepare(
#		qq{ DELETE FROM contributor WHERE id = ? }
#	);
#	$sth->execute($scrap_id);
#	$sth->finish;

	## Right now, we're going on the assumption that contributor data is only
	## added via this interface.
	my $sth = $db->prepare(qq{
		INSERT INTO contributor (scrap_id,name,email,contrib_date,data) VALUES(?,?,?,NOW(),?)
	});
	foreach my $k (@$contributors) {
		$sth->execute($scrap_id, $k->{name}, $k->{email}, $k->{note});
	}
	$sth->finish;
}

1;
