#!/usr/bin/perl -w
##
## Prototype Scraps server
## Copyright 2001 Dave O'Neill <dmo@acm.org>
##

use Frontier::RPC2;
use Digest::MD5 qw(md5_hex);
use HTTP::Daemon;
use HTTP::Status;
use DBI;

use Scraps;
use Scraps::Errors;
use Scraps::Config;
use Scraps::User;

## Advance prototype.
sub parse_request($$$$);


## Connect to database
my $dsn = "DBI:mysql:database=${Scraps::Config::database_name}:host=${Scraps::Config::database_host}";
my $db = DBI->connect($dsn, $Scraps::Config::database_user, $Scraps::Config::database_pass, { AutoCommit => 1 });
die("Error establishing database connect: $DBI::errstr\n") unless defined($db);
$db->{RaiseError} = 1;	## Die on errors.  

## We must remember to disconnect on exit.  Use the magical END
END { $db->disconnect() if defined($db); }

## Create server
my $srv = new HTTP::Daemon LocalPort => $Scraps::Config::rpc_port, Reuse => 1;

if(!defined($srv)) {
	print STDERR "Couldn't bind to port 8000.  Exiting\n";
	exit;
}

## Create XML-RPC decoder
my $rpc = new Frontier::RPC2 'use_objects' => 1;

## Create method mapping.
## FIXME: this should use something in the Scraps package.
## Or, preferably something Factory-like that can create
## objects as appropriate.
my $methods = {

	## Scrap handlers.
	'scraps.newScrap' => \&Scraps::newScrap,
	'scraps.fetchScrap' => \&Scraps::fetchScrap,
	'scraps.saveScrap' => \&Scraps::saveScrap,
	'scraps.search' => \&Scraps::search,
	'scraps.exportScrap' => \&Scraps::exportScrap,
	'scraps.deleteScrap' => \&Scraps::deleteScrap,
	'scraps.exportSearch' => \&Scraps::exportSearch,
	'scraps.import' => \&Scraps::importScrapbook,

	## User creation/deletion/verification/passwd-change handlers
	'scraps.user.verify' => \&Scraps::User::verify,
	'scraps.user.list' => \&Scraps::User::list,
	'scraps.user.add' => \&Scraps::User::add,
	'scraps.user.remove' => \&Scraps::User::remove,
	'scraps.user.changePassword' => \&Scraps::User::changePassword,

	
	 echo => sub { return [@_]; }
};

## FIXME: Only single-threaded right now.  Deal with forking later.
my $conn;

## Set up canned response object.  This can be safely reused in
## the connection loop.
my $response = new HTTP::Response 200;
$response->header('Content-Type' => 'text/xml');

print STDERR "Ready to accept requests\n";
while ($conn = $srv->accept) {
	my $rq = $conn->get_request;
	print STDERR "Got a request\n";
	print STDERR $rq->content, "\n";
	if($rq) {
		## XXX: Should the path be /RPC2, or is this just a holdover? 
		if ($rq->method eq 'POST' && $rq->url->path eq '/RPC2') {
			## We're OK.  Handle request
			print STDERR "Accepting request for ",$rq->url->path,"\n";
			$response->content(
				## FIXME: OK, the right way to do this might be to 
				## subclass Frontier::RPC2 and implement parse_request()
				## in the subclass?
				parse_request($rpc, $db, $methods, $rq->content)
			);
			$conn->send_response($response);
		} else {
			print STDERR "Rejecting request for ",$rq->url->path,"\n";
			$conn->send_error(RC_FORBIDDEN);
		}
	}
	$response->content('');
	$conn = undef;	## Forced close of connection.
}
		

##  This duplicates the code in Frontier::RPC2::serve(), but makes
##  it less sucky.  YMMV.
sub parse_request($$$$) 
{
	my $rpc = shift;
	my $db = shift;
	my $methods = shift;
	my $xml = shift;

	my $call;

	eval { $call = $rpc->decode($xml) };
	if ($@) {
		print STDERR "error decoding RPC\n";
		return $rpc->encode_fault(Scraps::INTERNAL_ERROR, "error decoding RPC.\n" . $@);
	}

	if ($call->{'type'} ne 'call') {
		print STDERR "error, got $call->{'type'} instead of methodCall \n";
		return $rpc->encode_fault(Scraps::INTERNAL_ERROR, "expected RPC \`methodCall', got \`$call->{'type'}'\n");
	}	

	## First off, we want to see if this user can do this method, so
	## the underlying code doesn't have to deal with it.
	## FIXME: check the database to see if:
	## 	-  user is valid, and/or password is valid.
	## Password is MD5'ed in the db, so we want to MD5 it here before
	## querying.  I suppose we could use MySQL's builtin md5() op, 
	## but...

	## Username/password are always the first two args.  
	my $username = shift @{$call->{'value'}} || '';
	my $password = shift @{$call->{'value'}} || '';
	
	## Sanity check on $username, Just In Case.
	$username =~ s/[^0-9a-zA-Z_. ]//;

	my $pw_hash  = md5_hex($password);

	## Now, validate user/password.
	my $sth = $db->prepare("SELECT username from user where username = ? AND password = ?");
	$sth->execute($username,$pw_hash) || die $sth->errstr;
	my $ar = $sth->fetchrow_arrayref;

	if(!defined($ar))
	{
		## FIXME: Return appropriate error code!
		print STDERR "error, got invalid username or password\n";
		return $rpc->encode_fault(Scraps::INVALID_AUTHENTICATION, "Invalid username or password.\n");
	}

	## Now check to see that the method exists.
	## We don't do this earlier, so that unauthorized users can't 
	## try random method calls to see if they work.
	my $method = $call->{'method_name'};
	if (!defined $methods->{$method}) {
		print STDERR "error, calling invalid method\n";
		return $rpc->encode_fault(Scraps::COMMAND_NOT_IMPLEMENTED, "no such method \`$method'\n");
	}
	
	## We need/want to pass the db handle along.  
	## We also assume that by this point, $username is valid.
	## This means $username has permission to view all un-grouped data,
	## and can view data in any group it's on the list for.
	my ($rv,$result);
	eval { ($rv,$result) = $methods->{$method}($username,$db,@{$call->{'value'}}) };
	if ($@) {
		print STDERR "error, didn't execute RPC call: $@\n";
		return $rpc->encode_fault(Scraps::INTERNAL_ERROR, "error executing RPC \`$method'.\n" . $@);
	}
	
	## If we have a nonzero return value, the $result is a text string
	## appropriate to the error.
	## If $rv == 0, $result is the hashref to the return value.
	print STDERR "Got return value: [$rv]\n";
	if ($rv != Scraps::SUCCESS) {
		print STDERR "Encoding fault: [$rv],[$result]\n";
		return $rpc->encode_fault($rv,$result);
	} else {
		if( $result =~ /^\d+/ && ($result == 1 || $result == 0) ) {
			## Properly encode boolean.  
			## We perform this hack here to keep underlying code
			## from needing any info about the XML-RPC protocol.
			$result = $rpc->boolean($result);
		}
		return $rpc->encode_response($result);
	}
}
